#include "MeshSimp.h"

BaseMeshSimplifier::BaseMeshSimplifier(ELEV **demData, OutFile *ivfile, int aDim)
{
  notRemovedCounter = 0;
  data = demData;
  dim = aDim;
  vrmlfile = ivfile;
  NewVector(datax, dim*dim, COOR);
  NewVector(dataz, dim*dim, COOR);
};

BaseMeshSimplifier::~BaseMeshSimplifier()
{
  DeleteVector(datax);
  DeleteVector(dataz);
};

void
BaseMeshSimplifier::simplifyMesh(ELEV thresh)
{
  fatal("Abstract mesh simplification!");
};

void 
BaseMeshSimplifier::outputPoint(COOR i, COOR j)
{
  vrmlfile->writeNode(dim - j - 1, i, data[i][j]);
  datax[++notRemovedCounter] = i;
  dataz[notRemovedCounter] = j;
};

int
BaseMeshSimplifier::getNotRemovedCounter()
{
  return notRemovedCounter;
};


//////////////////////////////////////////////////
MeshSimplifier::MeshSimplifier(ELEV **demData, OutFile *ivfile, int aDim):
  BaseMeshSimplifier(demData, ivfile, aDim)
{
  prepareAdditionalData();
}

MeshSimplifier::~MeshSimplifier()
{
  removeAdditionalData();
}
  
void
MeshSimplifier::simplifyMesh(ELEV thresh)
{
  info("Mesh-simplification...", dim);
  COOR i, j, dir, xp;
  ELEV y0, yp, yn;
  float cp;
  boolean removed;

  for(i=1; i<dim-1; i++)
    for(j=1; j<dim-1; j++) {

      removed = FALSE;
      y0 = data[i][j];
      
      dir = 0;
      do {
	if (!removed) {
	  yn = data[i+dx[dir]][j+dy[dir]];
	  xp = adddata[i-dx[dir]][j-dy[dir]][dir] + 2;
	  yp = data[i-dx[dir]*(xp-1)][j-dy[dir]*(xp-1)];
	  cp = crossProductOf1dx(-yn+y0, xp, yn-yp);
	  if (fabs(1.0*cp)<thresh) {
	    removed = TRUE;
	    adddata[i][j][dir] = xp-1;
	  } 
	} else   //the node is removed, so update the distances
	  adddata[i][j][dir] = adddata[i-dx[dir]][j-dx[dir]][dir] + 1;
	dir++;
      } while (dir<4);  //&& !removed
      if (!removed) {
	adddata[i][j][0] = adddata[i][j][1] = 
	  adddata[i][j][2] = adddata[i][j][3] = 0;
	outputPoint(i, j);
      }
    }
  //boundaries
  for(i=1; i<dim-1; i++) {
    y0 = data[i][0];
    yn = data[i+1][0];
    xp = adddata[i-1][0][2]+2; //the direction is 2 - down
    yp = data[i-(xp-1)][0];
    cp = crossProductOf1dx(-yn+y0, xp, yn-yp);
    if (fabs(1.0*cp)<thresh) 
      adddata[i][0][2] = xp-1;
    else {
      adddata[i][0][2] = 0;
      outputPoint(i, 0);
    }
    y0 = data[i][dim-1];
    yn = data[i+1][dim-1];
    xp = adddata[i-1][dim-1][2]+2; //the direction is 2 - down
    yp = data[i-(xp-1)][dim-1];
    cp = crossProductOf1dx(-yn+y0, xp, yn-yp);
    if (fabs(1.0*cp)<thresh) 
      adddata[i][dim-1][2] = xp-1;
    else {
      adddata[i][dim-1][2] = 0;
      outputPoint(i, dim-1);
    }
  }

  for(j=1; j<dim-1; j++) {
    y0 = data[0][j];
    yn = data[0][j+1];
    xp = adddata[0][j-1][0]+2; //the direction is 0 - right
    yp = data[0][j-(xp-1)];
    cp = crossProductOf1dx(-yn+y0, xp, yn-yp);
    if (fabs(1.0*cp)<thresh) 
      adddata[0][j][0] = xp-1;
    else {
      adddata[0][j][0] = 0;
      outputPoint(0, j);
    }
    y0 = data[dim-1][j];
    yn = data[dim-1][j+1];
    xp = adddata[dim-1][j-1][0]+2; //the direction is 0 - right
    yp = data[dim-1][j-(xp-1)];
    cp = crossProductOf1dx(-yn+y0, xp, yn-yp);
    if (fabs(1.0*cp)<thresh) 
      adddata[dim-1][j][0] = xp-1;
    else {
      adddata[dim-1][j][0] = 0;
      outputPoint(dim-1, j);
    }
  }

  // the corners are never removed!!
  outputPoint(0, 0);
  outputPoint(0, dim-1);
  outputPoint(dim-1, dim-1);
  outputPoint(dim-1, 0);
  info("...is done\n");
  
}

void 
MeshSimplifier::prepareAdditionalData()
{
  int i, j;
  NewVector(adddata, dim, COOR **);
  for(i=0; i<dim; i++) {
    NewVector(adddata[i], dim, COOR*);
    for(j=0; j<dim; j++) {
      NewVector(adddata[i][j], 4, COOR);
      adddata[i][j][0] = adddata[i][j][1] = 
	adddata[i][j][2] = adddata[i][j][3] = 0;
    }
  }
}

void 
MeshSimplifier::removeAdditionalData()
{
  int i, j;
  for(i=0; i<dim; i++) {
    for(j=0; j<dim; j++)
      DeleteVector(adddata[i][j]);
    DeleteVector(adddata[i]);
  }
  DeleteVector(adddata);
}

///////////////////////////////////

SecDerSimplifier::SecDerSimplifier(ELEV **demData, OutFile *ivfile, int aDim):
  BaseMeshSimplifier(demData, ivfile, aDim)
{
  prepareDerData();
}

SecDerSimplifier::~SecDerSimplifier()
{
  removeDerData();
}

void 
SecDerSimplifier::simplifyMesh(ELEV thresh)
{
  int i, j;
  
  for(i=1; i<dim-1; i++)
    for(j=1; j<dim-1; j++)
      if(fabs(secDerData1[i-1][j-1]) + 
	 fabs(secDerData2[i-1][j-1]) +
	 fabs(secDerData3[i-1][j-1]) +
	 fabs(secDerData3[i-1][j-1])>=thresh)
	outputPoint(i,j);

  // edges
  for(i=1; i<dim-2; i++) {
    ELEV tmp1 = fabs(data[0][i] - data[0][i-1]);
    ELEV tmp2 = fabs(data[0][i+1] - data[0][i]);
    if (tmp1+tmp2>=thresh/2.0)
      outputPoint(0, i);
    tmp1 = fabs(data[dim-1][i] - data[dim-1][i-1]);
    tmp2 = fabs(data[dim-1][i+1] - data[dim-1][i]);
    if (tmp1+tmp2>=thresh/2.0)
      outputPoint(dim-1, i);
  }
  
  for(i=1; i<dim-2; i++) {
    ELEV tmp1 = fabs(data[i][0] - data[i-1][0]);
    ELEV tmp2 = fabs(data[i+1][0] - data[i][0]);
    if (tmp1+tmp2>=thresh/2.0)
      outputPoint(i, 0);
    tmp1 = fabs(data[i][dim-1] - data[i-1][dim-1]);
    tmp2 = fabs(data[i+1][dim-1] - data[i][dim-1]);
    if (tmp1+tmp2>=thresh/2.0)
      outputPoint(i, dim-1);
  }

  outputPoint(0,0);
  outputPoint(dim-1,0);
  outputPoint(dim-1,dim-1);
  outputPoint(0,dim-1);
};

void
SecDerSimplifier::prepareDerData()
{
  info("in prepare derdata");
  
  int i,j;
  NewMatrix(derData1, dim-1, dim-1, ELEV);
  NewMatrix(secDerData1, dim-2, dim-2, ELEV);
  NewMatrix(derData2, dim-1, dim-1, ELEV);
  NewMatrix(secDerData2, dim-2, dim-2, ELEV);
  NewMatrix(derData3, dim-1, dim-1, ELEV);
  NewMatrix(secDerData3, dim-2, dim-2, ELEV);
  NewMatrix(derData4, dim-1, dim-1, ELEV);
  NewMatrix(secDerData4, dim-2, dim-2, ELEV);
  
  //first derivative
  for(i=0; i<dim-1; i++)
    for(j=0; j<dim-1; j++) {
      derData1[i][j] = data[i+1][j+1] - data[i][j];  //xy direction
      derData2[i][j] = data[i+1][j] - data[i][j+1];  //-xy direction
      derData3[i][j] = (data[i][j+1]+data[i+1][j+1])/2 - 
	(data[i][j]+data[i+1][j])/2; //x direction
      derData4[i][j] = (data[i+1][j+1]+data[i+1][j])/2 - 
	(data[i][j]+data[i][j+1])/2; //y direction
    }
  
  //second one
  for(i=0; i<dim-2; i++)
    for(j=0; j<dim-2; j++) {
      secDerData1[i][j] = derData1[i+1][j+1] - derData1[i][j];  //xy direction
      secDerData2[i][j] = derData2[i+1][j] - derData2[i][j+1];  //-xy direction
      secDerData3[i][j] = (derData3[i][j+1]+derData3[i+1][j+1])/2 - 
	(derData3[i][j]+derData3[i+1][j])/2; //x direction
      secDerData4[i][j] = (derData4[i+1][j+1]+derData4[i+1][j])/2 - 
	(derData4[i][j]+derData4[i][j+1])/2; //y direction
    }
};

void
SecDerSimplifier::removeDerData()
{
  DeleteMatrix(derData1);
  DeleteMatrix(secDerData1);
  DeleteMatrix(derData2);
  DeleteMatrix(secDerData2);
  DeleteMatrix(derData3);
  DeleteMatrix(secDerData3);
  DeleteMatrix(derData4);
  DeleteMatrix(secDerData4);
};
